//! Data structures and operations for game inputs, particularly ones that induce
//! movements to the current piece.

use crate::matrix::Mat;
use crate::piece::{Piece, Rot, Shape};

/// Represents a basic movement input from the user.
///
/// DAS inputs are not included, since functionally DAS is equivalent to repeated taps,
/// and a bot can effectively hypertap infinitely fast during its analysis.
#[derive(Copy, Clone, Eq, PartialEq, Ord, PartialOrd, Debug, Hash)]
pub enum Movement {
    /// "Sonic" drop.
    Drop,
    /// Horizontal shift.
    Shift(Horizontal),
    /// Rotation.
    Rotate(Spin),
    // TODO: soft drops?
}

impl Movement {
    pub const DROP: Movement = Movement::Drop;
    pub const LEFT: Movement = Movement::Shift(Horizontal::Left);
    pub const RIGHT: Movement = Movement::Shift(Horizontal::Right);
    pub const CW: Movement = Movement::Rotate(Spin::Cw);
    pub const CCW: Movement = Movement::Rotate(Spin::Ccw);
    pub const FLIP: Movement = Movement::Rotate(Spin::Flip);

    /// Applies this movement to the given piece on the given matrix. Returns true if the
    /// movement was successful, or false if the piece did not move because it was
    /// obstructed by the matrix. `piece` will be modified in-place iff the return value
    /// is true.
    pub fn perform<'c, 'k, S>(&self, piece: &mut Piece<S>, mat: &Mat) -> bool
    where
        S: Shape<'c> + Kicks<'k>,
    {
        match *self {
            Self::Drop => drop(piece, mat),
            Self::Shift(d) => shift(piece, mat, d),
            Self::Rotate(d) => rotate(piece, mat, d).is_some(),
        }
    }
}

/// Represents a direction of horizontal movement.
#[derive(Copy, Clone, Eq, PartialEq, Ord, PartialOrd, Debug, Hash)]
#[repr(i16)]
pub enum Horizontal {
    /// Left.
    Left = -1,
    /// Right.
    Right = 1,
}

/// Represents a rotating operation. Includes 180 degree "flips", which are non-standard
/// for tetris but exist in many unofficial iterations of the game.
///
/// Rotations may be performed on [rotation states](Rot) using the [addition
/// operator](core::ops::Add).
///
/// ```
/// # use mino::{piece::Rot, input::Spin};
/// assert_eq!(Rot::N + Spin::Cw, Rot::E);
/// ```
#[derive(Copy, Clone, Eq, PartialEq, Ord, PartialOrd, Debug, Hash)]
#[repr(i8)]
pub enum Spin {
    /// Clockwise.
    Cw = 1,
    /// 180 degrees.
    Flip = 2,
    /// Counterclockwise.
    Ccw = -1,
}

impl core::ops::Add<Spin> for Rot {
    type Output = Rot;
    fn add(self, r: Spin) -> Rot {
        (self as i8).wrapping_add(r as i8).into()
    }
}

impl core::ops::AddAssign<Spin> for Rot {
    fn add_assign(&mut self, r: Spin) {
        *self = *self + r;
    }
}

/// Interface for shapes that have a table of "kicks" to perform when checking if a
/// rotation is valid.
pub trait Kicks<'k> {
    /// Returns the list of offsets to test when rotating from `r0` to `r1`. Offsets are
    /// tested from left to right. If `r0` == `r1`, then this function is permitted to
    /// return any list (incl. empty list) since that condition will never occur in
    /// practice.
    ///
    /// Note: `(0,0)` is not tested by default so it must be explicitly included if an
    /// immobile test should be performed. In this implementation of SRS, the O piece
    /// moves around an origin so its center must be corrected by a singular "kick" for
    /// each rotation, thus the O piece never tests `(0,0)` during a rotation.
    fn kicks(&self, r0: Rot, r1: Rot) -> &'k [(i16, i16)];
}

/// Drops `piece` until it hits `mat`. Returns true if the piece fell at least one block.
pub fn drop<'c, S>(piece: &mut Piece<S>, mat: &Mat) -> bool
where
    S: Shape<'c>,
{
    let mut cells = piece.cells();
    let mut did_fall = false;

    // XXX(iitalics): optimization: if piece is above the top of the matrix, then
    // immediately drop it to the row containing data since all space above that is
    // guarunteed unoccupied.
    let y0 = cells.extents().1.start;
    if y0 > mat.rows() {
        let dy = mat.rows() - y0;
        cells = cells.translate(0, dy);
        piece.loc.y += dy;
        did_fall = true;
    }

    loop {
        cells = cells.translate(0, -1);
        if cells.intersects(mat) {
            break;
        }
        piece.loc.y -= 1;
        did_fall = true;
    }
    did_fall
}

/// Moves a piece horizontally. Returns true if it successfully shifted the piece.
pub fn shift<'c, S>(piece: &mut Piece<S>, mat: &Mat, dir: Horizontal) -> bool
where
    S: Shape<'c>,
{
    let dx = dir as i16;
    if !piece.cells().translate(dx, 0).intersects(mat) {
        piece.loc.x += dx;
        return true;
    }
    false
}

/// Rotates a piece, performing kicks to find a final location. Returns `Some(idx)` if
/// successful, where `idx` is the index into the kick table performed. Index `0` means it
/// was not kicked. Returns `None` if the rotation failed.
pub fn rotate<'c, 'k, T>(piece: &mut Piece<T>, mat: &Mat, dir: Spin) -> Option<usize>
where
    T: Shape<'c> + Kicks<'k>,
{
    let r0 = piece.loc.r;
    let r1 = piece.loc.r + dir;
    let kicks = piece.ty.kicks(r0, r1);

    piece.loc.r = r1;
    let cells = piece.cells();

    for (i, &(dx, dy)) in kicks.iter().enumerate() {
        if !cells.translate(dx, dy).intersects(mat) {
            piece.loc.x += dx;
            piece.loc.y += dy;
            return Some(i);
        }
    }

    piece.loc.r = r0;
    None
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::mat;
    use crate::piece::test::Tri;
    use crate::piece::Loc;

    #[test]
    fn test_angle_rotate() {
        use Rot::*;
        use Spin::*;
        assert_eq!(N + Cw, E);
        assert_eq!(N + Ccw, W);
        assert_eq!(N + Flip, S);
        assert_eq!(E + Cw, S);
        assert_eq!(E + Ccw, N);
        assert_eq!(E + Flip, W);
        assert_eq!(S + Cw, W);
        assert_eq!(S + Ccw, E);
        assert_eq!(S + Flip, N);
        assert_eq!(W + Cw, N);
        assert_eq!(W + Ccw, S);
        assert_eq!(W + Flip, E);

        let mut r = N;
        r += Ccw;
        assert_eq!(r, W);
    }

    const MAT: &Mat = mat! {
        ".........x";
        "..x.....xx";
        "xxx.xxx.xx";
    };

    fn drop(x: i16, y: i16, r: Rot) -> i16 {
        let mut piece = Piece {
            ty: Tri,
            loc: Loc { x, y, r },
        };
        let did_fall = super::drop(&mut piece, MAT);
        assert_eq!(piece.loc.x, x);
        assert_eq!(did_fall, piece.loc.y != y, "{},{}", piece.loc.y, y);
        piece.loc.y
    }

    #[test]
    fn test_drop() {
        assert_eq!(drop(3, 2, Rot::N), 1);
        assert_eq!(drop(3, 8, Rot::N), 1);
        assert_eq!(drop(3, 1, Rot::N), 1);
        assert_eq!(drop(3, 8, Rot::E), 1);
        assert_eq!(drop(3, 1, Rot::E), 1);
        assert_eq!(drop(4, 8, Rot::E), 2);
        assert_eq!(drop(4, 2, Rot::E), 2);
        assert_eq!(drop(3, 8, Rot::S), 2);
        assert_eq!(drop(3, 2, Rot::S), 2);
        assert_eq!(drop(9, 8, Rot::W), 3);
        assert_eq!(drop(9, 4, Rot::W), 3);
        assert_eq!(drop(9, 3, Rot::W), 3);
        assert_eq!(drop(9, 8, Rot::S), 4);
        assert_eq!(drop(9, 5, Rot::S), 4);
        assert_eq!(drop(9, 4, Rot::S), 4);
    }

    fn shift(x: i16, y: i16, r: Rot, dir: Horizontal) -> i16 {
        let mut piece = Piece {
            ty: Tri,
            loc: Loc { x, y, r },
        };
        let did_shift = super::shift(&mut piece, MAT, dir);
        assert_eq!(piece.loc.y, y);
        assert_eq!(did_shift, piece.loc.x != x, "{},{}", piece.loc.x, x);
        piece.loc.x
    }

    #[test]
    fn test_shift() {
        use Horizontal::*;
        assert_eq!(shift(0, 1, Rot::N, Left), 0);
        assert_eq!(shift(0, 1, Rot::N, Right), 0);
        assert_eq!(shift(0, 2, Rot::N, Left), 0);
        assert_eq!(shift(0, 2, Rot::N, Right), 1);
        assert_eq!(shift(1, 2, Rot::N, Left), 0);
        assert_eq!(shift(1, 2, Rot::N, Right), 2);
        assert_eq!(shift(3, 1, Rot::E, Left), 3);
        assert_eq!(shift(3, 1, Rot::E, Right), 3);
    }

    impl Kicks<'static> for Tri {
        fn kicks(&self, r0: Rot, r1: Rot) -> &'static [(i16, i16)] {
            match (r0, r1) {
                (Rot::N, Rot::E) => &[(0, 0), (0, 1)],
                (Rot::N, Rot::W) => &[(0, 0), (1, 0)],
                (Rot::E, Rot::S) => &[(0, 0), (1, 0)],
                (Rot::E, Rot::N) => &[(0, 0), (0, -1)],
                (Rot::S, Rot::W) => &[(0, 0), (0, -1)],
                (Rot::S, Rot::E) => &[(0, 0), (-1, 0)],
                (Rot::W, Rot::N) => &[(0, 0), (-1, 0)],
                (Rot::W, Rot::S) => &[(0, 0), (0, 1)],
                (_, _) => &[(0, 0)], // flip or non rotation
            }
        }
    }

    fn rotate(x: i16, y: i16, r: Rot, dir: Spin) -> (i16, i16, Rot, Option<usize>) {
        let loc = Loc { x, y, r };
        let mut piece = Piece { ty: Tri, loc };
        let result = super::rotate(&mut piece, MAT, dir);
        assert_eq!(result.is_some(), piece.loc != loc, "{piece:?},{loc:?}");
        (piece.loc.x, piece.loc.y, piece.loc.r, result)
    }

    #[test]
    fn test_rotate() {
        use Rot::*;
        use Spin::*;
        assert_eq!(rotate(3, 2, N, Cw), (3, 2, E, Some(0)));
        assert_eq!(rotate(3, 2, E, Cw), (3, 2, S, Some(0)));
        assert_eq!(rotate(3, 2, S, Cw), (3, 2, W, Some(0)));
        assert_eq!(rotate(3, 2, W, Cw), (3, 2, N, Some(0)));
        assert_eq!(rotate(3, 2, N, Ccw), (3, 2, W, Some(0)));
        assert_eq!(rotate(3, 2, W, Ccw), (3, 2, S, Some(0)));
        assert_eq!(rotate(3, 2, S, Ccw), (3, 2, E, Some(0)));
        assert_eq!(rotate(3, 2, E, Ccw), (3, 2, N, Some(0)));
        assert_eq!(rotate(7, 1, S, Cw), (7, 1, W, Some(0)));
        assert_eq!(rotate(7, 1, S, Ccw), (7, 1, S, None)); // locked in
        assert_eq!(rotate(0, 1, N, Ccw), (1, 1, W, Some(1))); // kicks right
        assert_eq!(rotate(0, 1, N, Cw), (0, 2, E, Some(1))); // kicks up
    }
}
