//! Implementation of SRS shapes and movement quirks.

use crate::input::Kicks;
use crate::piece::{Cells, Rot, Shape, Spawn};

mod code_gen {
    include!(concat!(env!("OUT_DIR"), "/srs.rs"));
}

#[derive(Copy, Clone, Eq, PartialEq, Ord, PartialOrd, Debug, Hash)]
#[repr(u8)]
pub enum PieceType {
    I = code_gen::shape_indices::I as u8,
    J = code_gen::shape_indices::J as u8,
    L = code_gen::shape_indices::L as u8,
    O = code_gen::shape_indices::O as u8,
    S = code_gen::shape_indices::S as u8,
    T = code_gen::shape_indices::T as u8,
    Z = code_gen::shape_indices::Z as u8,
}

impl Shape<'static> for PieceType {
    fn cells(&self, r: Rot) -> Cells<'static> {
        let i = (*self as usize) * 4 + r as usize;
        let xs = code_gen::extents::X0[i]..code_gen::extents::X1[i];
        let ys = code_gen::extents::Y0[i]..code_gen::extents::Y1[i];
        let idx = code_gen::CELLS_INDEX[i];
        let data = &code_gen::CELLS_BITS[idx..idx + ys.len()];
        Cells::new(xs, ys, data)
    }
}

impl Kicks<'static> for PieceType {
    fn kicks(&self, r0: Rot, r1: Rot) -> &'static [(i16, i16)] {
        let i = (*self as usize) * 16 + (r0 as usize) * 4 + r1 as usize;
        let idx = code_gen::KICKS_INDEX[i];
        let len = code_gen::KICKS_COUNT[i];
        &code_gen::KICKS_TESTS[idx..idx + len]
    }
}

impl Spawn for PieceType {
    fn spawn_pos(&self) -> (i16, i16) {
        code_gen::SPAWN
    }
}

impl PieceType {
    /// Returns the name of the piece as a string.
    pub fn name(self) -> &'static str {
        match self {
            PieceType::I => "I",
            PieceType::J => "J",
            PieceType::L => "L",
            PieceType::O => "O",
            PieceType::S => "S",
            PieceType::T => "T",
            PieceType::Z => "Z",
        }
    }

    /// Returns the name of the piece as a [`char`].
    pub fn as_char(self) -> char {
        match self {
            PieceType::I => 'I',
            PieceType::J => 'J',
            PieceType::L => 'L',
            PieceType::O => 'O',
            PieceType::S => 'S',
            PieceType::T => 'T',
            PieceType::Z => 'Z',
        }
    }
}

impl core::fmt::Display for PieceType {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.write_str(self.name())
    }
}

/// Error returned when trying to convert an invalid string to a [`PieceType`].
#[derive(Debug)]
pub struct InvalidPieceName;

impl TryFrom<char> for PieceType {
    type Error = InvalidPieceName;
    fn try_from(ch: char) -> Result<PieceType, InvalidPieceName> {
        match ch {
            'I' => Ok(PieceType::I),
            'J' => Ok(PieceType::J),
            'L' => Ok(PieceType::L),
            'O' => Ok(PieceType::O),
            'S' => Ok(PieceType::S),
            'T' => Ok(PieceType::T),
            'Z' => Ok(PieceType::Z),
            _ => Err(InvalidPieceName),
        }
    }
}

impl core::str::FromStr for PieceType {
    type Err = InvalidPieceName;
    fn from_str(s: &str) -> Result<PieceType, InvalidPieceName> {
        if s.len() != 1 {
            return Err(InvalidPieceName);
        }
        s.chars().next().unwrap().try_into()
    }
}

impl core::fmt::Display for InvalidPieceName {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(f, "invalid piece name")
    }
}

pub type Piece = crate::piece::Piece<PieceType>;
pub type Queue<'a> = crate::queue::Queue<'a, PieceType>;

#[cfg(test)]
mod test {
    use super::*;
    use crate::matrix::Mat;
    use crate::piece::Loc;
    use crate::{input, mat};

    use alloc::vec::Vec;
    use core::ops::Range;

    use PieceType::*;

    #[test]
    fn test_spawn() {
        for ty in [L, O, J, I, S, T, Z] {
            assert_eq!(Piece::spawn(ty).loc, (4, 19, Rot::N).into());
        }
    }

    #[test]
    fn test_piece_name_parse() {
        for ty in [L, O, J, I, S, T, Z] {
            assert_eq!(PieceType::try_from(ty.as_char()).unwrap(), ty);
            assert_eq!(ty.name().parse::<PieceType>().unwrap(), ty);
        }
    }

    fn get_cells(ty: PieceType, r: Rot) -> (Range<i16>, Range<i16>, Vec<(i16, i16)>) {
        let loc = Loc { x: 0, y: 0, r };
        let piece = Piece { ty, loc };
        let mut coords = piece.cells().coords().collect::<Vec<_>>();
        coords.sort();
        let (xs, ys) = piece.cells().extents();
        (xs, ys, coords)
    }

    fn compute_cells(init: &[(i16, i16)], r: Rot) -> (Range<i16>, Range<i16>, Vec<(i16, i16)>) {
        let mut coords = init.iter().map(|xy| rotate(*xy, r)).collect::<Vec<_>>();
        coords.sort();
        let min_x = coords.iter().map(|&(x, _)| x).min().unwrap();
        let max_x = coords.iter().map(|&(x, _)| x).max().unwrap();
        let min_y = coords.iter().map(|&(_, y)| y).min().unwrap();
        let max_y = coords.iter().map(|&(_, y)| y).max().unwrap();
        let xs = min_x..max_x + 1;
        let ys = min_y..max_y + 1;
        (xs, ys, coords)
    }

    fn rotate(mut xy: (i16, i16), r: Rot) -> (i16, i16) {
        for _ in 0..(r as u8) {
            let (x, y) = xy;
            xy = (y, -x);
        }
        xy
    }

    fn test_cells_all_rotations(ty: PieceType, expected_coords: &[(i16, i16)]) {
        for r in (0..4).map(Rot::from) {
            let (act_xs, act_ys, act_coords) = get_cells(ty, r);
            let (exp_xs, exp_ys, exp_coords) = compute_cells(expected_coords, r);
            assert_eq!(act_xs, exp_xs, "{ty:?},{r:?}");
            assert_eq!(act_ys, exp_ys, "{ty:?},{r:?}");
            assert_eq!(act_coords, exp_coords, "{ty:?},{r:?}");
        }
    }

    #[test]
    fn test_srs_cells() {
        test_cells_all_rotations(I, &[(-1, 0), (0, 0), (1, 0), (2, 0)]);
        test_cells_all_rotations(J, &[(-1, 1), (-1, 0), (0, 0), (1, 0)]);
        test_cells_all_rotations(L, &[(1, 1), (-1, 0), (0, 0), (1, 0)]);
        test_cells_all_rotations(O, &[(0, 0), (1, 0), (0, 1), (1, 1)]);
        test_cells_all_rotations(S, &[(-1, 0), (0, 0), (0, 1), (1, 1)]);
        test_cells_all_rotations(T, &[(0, 1), (-1, 0), (0, 0), (1, 0)]);
        test_cells_all_rotations(Z, &[(1, 0), (0, 0), (0, 1), (-1, 1)]);
    }

    #[test]
    fn test_srs_o_kicks_in_place() {
        use input::Spin::*;
        let mut piece = Piece {
            ty: O,
            loc: (6, 6, Rot::N).into(),
        };
        assert_eq!(input::rotate(&mut piece, Mat::EMPTY, Cw), Some(0));
        assert_eq!(piece.loc.x, 6);
        assert_eq!(piece.loc.y, 7);
        assert_eq!(input::rotate(&mut piece, Mat::EMPTY, Cw), Some(0));
        assert_eq!(piece.loc.x, 7);
        assert_eq!(piece.loc.y, 7);
        assert_eq!(input::rotate(&mut piece, Mat::EMPTY, Cw), Some(0));
        assert_eq!(piece.loc.x, 7);
        assert_eq!(piece.loc.y, 6);
        assert_eq!(input::rotate(&mut piece, Mat::EMPTY, Cw), Some(0));
        assert_eq!(piece.loc.x, 6);
        assert_eq!(piece.loc.y, 6);
    }

    #[test]
    fn test_srs_t_spin_triple() {
        const SETUP: &Mat = mat! {
            "......xx..";
            ".......x..";
            "xxxxxx.xxx";
            "xxxxx..xxx";
            "xxxxxx.xxx";
        };
        let mut piece = Piece {
            ty: T,
            loc: (5, 3, Rot::N).into(),
        };
        assert_eq!(input::rotate(&mut piece, SETUP, input::Spin::Ccw), Some(4));
        assert_eq!(piece.loc.x, 6);
        assert_eq!(piece.loc.y, 1);
        assert_eq!(piece.loc.r, Rot::W);
    }

    #[test]
    fn test_srs_z_spin_triple() {
        const SETUP: &Mat = mat! {
            "....x.....";
            "xxxxxx.xxx";
            "xxxxx..xxx";
            "xxxxx.xxxx";
        };
        let mut piece = Piece {
            ty: Z,
            loc: (5, 3, Rot::N).into(),
        };
        assert_eq!(input::rotate(&mut piece, SETUP, input::Spin::Cw), Some(3));
        assert_eq!(piece.loc.x, 5);
        assert_eq!(piece.loc.y, 1);
        assert_eq!(piece.loc.r, Rot::E);
    }

    #[test]
    fn test_srs_cells_eq() {
        fn rotate(ty: PieceType, dir: input::Spin) -> Piece {
            let mut piece = Piece::spawn(ty);
            input::rotate(&mut piece, Mat::EMPTY, dir).unwrap();
            piece
        }
        for ty in [S, Z, I] {
            let cw = rotate(ty, input::Spin::Cw);
            let ccw = rotate(ty, input::Spin::Ccw);
            // rotation states differ
            assert_ne!(cw.loc.r, ccw.loc.r, "ty={ty:?}");
            // cells() are equal after translating, since they occupy the same blocks
            assert_eq!(cw.cells().translate(-1, 0), ccw.cells(), "ty={ty:?}");
        }
    }

    #[test]
    fn test_o_cells_eq() {
        fn cells(x: i16, y: i16, r: Rot) -> Cells<'static> {
            Piece {
                ty: O,
                loc: Loc { x, y, r },
            }
            .cells()
        }
        assert_eq!(cells(0, 0, Rot::N), cells(0, 1, Rot::E));
        assert_eq!(cells(0, 0, Rot::N), cells(1, 1, Rot::S));
        assert_eq!(cells(0, 0, Rot::N), cells(1, 0, Rot::W));
    }
}
