use anyhow::{Context as _, Result};
use fish::bot::Bot;
use mino::queue::Queue;
use rand::RngCore as _;
use std::collections::HashMap;
use std::io::{Read, Write};
use std::path::Path;
use std::sync::{atomic, mpsc, Arc};
use std::time::{Duration, Instant};

use tidepool::cli::{parse_cli, IoArgs, Mode, MultiRun, OutputDataArgs, SingleRun};
use tidepool::config::Config;
use tidepool::output::{Move, Output, SummaryStats};
use tidepool::sim::{Simul, SimulOptions};

fn main() -> Result<()> {
    tracing_subscriber::fmt::fmt()
        .with_writer(std::io::stderr)
        .with_env_filter(tracing_subscriber::filter::EnvFilter::from_default_env())
        .init();

    match parse_cli() {
        Mode::Single(args) => single(args),
        Mode::Multi(args) => multi(args),
    }
}

// syncronize multiple threads by having them all send messages of this type to a common
// channel, which receives and process those messages sequentially on the main thread
#[derive(Debug)]
enum Msg {
    Interrupt,
    Heartbeat(usize),
    Shutdown(usize),
    Status(usize, Box<Status>),
    Output(usize, Box<Output>),
}

#[derive(Debug)]
struct Status {
    pieces: usize,
    lines_left: usize,
    time: Duration,
}

fn create_mailbox() -> (mpsc::SyncSender<Msg>, mpsc::Receiver<Msg>) {
    let (tx, rx) = mpsc::sync_channel(128);

    let tx2 = tx.clone();
    let on_ctrlc = move || {
        if tx2.send(Msg::Interrupt).is_err() {
            tracing::warn!("nobody handled ctrl-c, aborting");
            std::process::exit(128 + ctrlc::Signal::SIGINT as i32);
        }
    };
    if let Err(err) = ctrlc::set_handler(on_ctrlc) {
        tracing::warn!("could not set signal handler: {err}");
    }

    (tx, rx)
}

fn single(args: SingleRun) -> Result<()> {
    let config = read_config_file(&args.config_file)?;

    let (tx, rx) = create_mailbox();
    let exit_early = Arc::new(atomic::AtomicBool::new(false));

    std::thread::spawn({
        let config = config.clone();
        let exit_early = exit_early.clone();
        move || run_simulation(&args.data, &config, args.seed, 0, &tx, &exit_early)
    });

    while let Ok(msg) = rx.recv() {
        tracing::trace!(msg = debug(&msg));

        match msg {
            Msg::Interrupt => {
                eprint!("\ncaught interrupt.");
                exit_early.store(true, atomic::Ordering::Relaxed);
            }
            Msg::Heartbeat(_id) => {
                tracing::debug!("heartbeat");
            }
            Msg::Shutdown(_id) => {
                // only used by `run_simulations()`
                unreachable!()
            }
            Msg::Status(_id, status) => {
                if !exit_early.load(atomic::Ordering::Relaxed) {
                    print_single_progress(&args.io, &status, config.game.goal)?;
                }
            }
            Msg::Output(_id, output) => {
                // dropping rx early does two things here: 1) removing any of the 'break's
                // in this block become a compile error, 2) interrupts will not be able to
                // be sent so the ctrl-c handler will abort the program
                std::mem::drop(rx);

                eprintln!();

                if !output.did_complete() && !prompt_yn(&args.io, "run did not finish, keep it?") {
                    break;
                }

                write_output(&args.io, args.output_file.as_deref(), &output)?;
                break;
            }
        }
    }

    Ok(())
}

fn multi(args: MultiRun) -> Result<()> {
    let config = read_config_file(&args.config_file)?;

    let (tx, rx) = create_mailbox();
    let tasks = args.count.map_or(i64::MAX, |n| n as i64);
    let tasks = Arc::new(atomic::AtomicI64::new(tasks));
    let exit_early = Arc::new(atomic::AtomicBool::new(false));

    fn default_num_cpus() -> usize {
        match std::thread::available_parallelism() {
            Ok(n) => n.get(),
            Err(_) => 1,
        }
    }

    let n_jobs = args.jobs.unwrap_or_else(default_num_cpus);
    let mut jobs = HashMap::with_capacity(n_jobs);
    let mut completed = SummaryStats::<usize>::new();

    tracing::debug!("spawning {n_jobs} jobs");
    for id in 0..n_jobs {
        jobs.insert(id, None);
        std::thread::spawn({
            let data_args = args.data.clone();
            let config = config.clone();
            let tx = tx.clone();
            let tasks = tasks.clone();
            let exit_early = exit_early.clone();
            move || run_simulations(&data_args, &config, id, &tx, &tasks, &exit_early)
        });
    }

    while let Ok(msg) = rx.recv() {
        tracing::trace!(msg = debug(&msg));
        match msg {
            Msg::Interrupt => {
                eprint!("\ncaught interrupt.");
                tasks.store(0, atomic::Ordering::Relaxed);
                exit_early.store(true, atomic::Ordering::Relaxed);
            }
            Msg::Heartbeat(id) => {
                tracing::debug!("heartbeat {id}");
            }
            Msg::Shutdown(id) => {
                tracing::debug!("shutdown {id}");
                assert!(jobs.remove(&id).is_some(), "shutdown {id}");
                if jobs.is_empty() {
                    break;
                }
            }
            Msg::Status(id, status) => {
                *jobs.get_mut(&id).unwrap() = Some(status);

                if !exit_early.load(atomic::Ordering::Relaxed) {
                    print_multi_progress(&args.io, &jobs, &completed)?;
                }
            }
            Msg::Output(_id, output) => {
                if output.did_complete() {
                    completed.insert(output.pieces);
                }

                write_output_in_dir(&args.output_dir, &output)?;

                if !exit_early.load(atomic::Ordering::Relaxed) {
                    print_multi_progress(&args.io, &jobs, &completed)?;
                }
            }
        }
    }

    eprintln!();

    Ok(())
}

fn read_config_file(path: &Path) -> Result<Config> {
    enum Format {
        Json,
        Toml,
    }

    let extension = path.extension().and_then(|e| e.to_str());
    let format = match extension {
        Some(e) if e.eq_ignore_ascii_case("json") => Format::Json,
        Some(e) if e.eq_ignore_ascii_case("toml") => Format::Toml,
        _ => {
            return Err(anyhow::anyhow!(
                "invalid config file extension, supports .toml or .json"
            ));
        }
    };

    let mut contents = String::new();
    std::fs::File::open(path)
        .with_context(|| format!("error opening config file '{}'", path.display()))?
        .read_to_string(&mut contents)
        .context("error reading config file")?;

    match format {
        Format::Json => serde_json::from_str(&contents).context("error parsing config file"),
        Format::Toml => toml::from_str(&contents).context("error parsing config file"),
    }
}

fn write_output(io_args: &IoArgs, path: Option<&Path>, output: &Output) -> Result<()> {
    let mut writer: Box<dyn std::io::Write> = match &path {
        Some(path) => {
            if path.is_file() {
                if prompt_yn(io_args, "output file already exists, overwrite?") {
                    tracing::debug!("removing {path:?}");
                    std::fs::remove_file(path)?;
                } else {
                    tracing::warn!("output file already exists, exiting.");
                    return Ok(());
                }
            }
            Box::new(
                std::fs::File::create(path)
                    .with_context(|| format!("error opening output file '{}'", path.display()))?,
            )
        }
        None => Box::new(std::io::stdout()),
    };

    let pretty = path.is_none() && !io_args.quiet;
    if pretty {
        serde_json::to_writer_pretty(&mut writer, &output)
            .context("error writing output")
            .and_then(|_| writeln!(&mut writer).context("error writing output"))
    } else {
        serde_json::to_writer(&mut writer, &output).context("error writing output")
    }
}

fn write_output_in_dir(dir_path: &Path, output: &Output) -> Result<()> {
    let date: chrono::DateTime<chrono::Local> = std::time::SystemTime::now().into();

    std::fs::create_dir_all(dir_path).context("error creating directory to store output")?;

    let goal = output.config.goal;
    let pieces = output.pieces;
    let incomplete = if !output.did_complete() { "I-" } else { "" };
    let date = date.format("%Y-%m-%d_%H-%M-%S");
    let file_name = format!("{goal}L-{incomplete}{pieces}p-{date}.json");

    let io_args = IoArgs {
        // overwrite existing file
        noninteractive: true,
        // don't prompt about incomplete run
        quiet: true,
    };
    let path = dir_path.join(file_name);
    write_output(&io_args, Some(&path), output)
}

fn print_single_progress(io_args: &IoArgs, status: &Status, goal: usize) -> std::io::Result<()> {
    if io_args.quiet {
        return Ok(());
    }

    let width = 40; // TODO: get terminal size?
    let pieces = status.pieces;
    let cleared = goal - status.lines_left;
    let pps = pieces as f64 / status.time.as_secs_f64().max(0.001);
    let pace = pieces * 100 / cleared.max(1);

    let writer = std::io::stderr().lock();
    let mut writer = std::io::BufWriter::new(writer);
    write!(writer, "\r[")?;
    for i in 0..width {
        if goal * i <= cleared * (width - 1) {
            write!(writer, "#")
        } else {
            write!(writer, " ")
        }?;
    }
    write!(
        writer,
        "] {cleared}/{goal} #{pieces}, {pps:.2} pps, {pace} pace"
    )?;

    write!(writer, " ")?;
    writer.flush()
}

fn print_multi_progress(
    io_args: &IoArgs,
    jobs: &HashMap<usize, Option<Box<Status>>>,
    completed: &SummaryStats<usize>,
) -> std::io::Result<()> {
    if io_args.quiet {
        return Ok(());
    }

    let writer = std::io::stderr().lock();
    let mut writer = std::io::BufWriter::new(writer);
    write!(writer, "\rjobs:{}", jobs.len())?;

    let mut pps_stats = SummaryStats::new();
    for status in jobs.values().filter_map(Option::as_deref) {
        let pps = status.pieces as f64 / status.time.as_secs_f64().max(0.001);
        pps_stats.insert_f64(pps);
    }
    if let Some(pps) = pps_stats.avg_f64() {
        write!(writer, ", pps(avg):{pps:.2}")?;
    }

    if completed.count() > 0 {
        let count = completed.count();
        let min = completed.min().unwrap();
        let max = completed.max().unwrap();
        let avg = completed.avg().unwrap();
        write!(
            writer,
            ", completed:{count}, min:#{min}, max:#{max}, avg:#{avg}"
        )?;
    }

    write!(writer, " ")?;
    writer.flush()
}

fn prompt_yn(io_args: &IoArgs, msg: &str) -> bool {
    if io_args.noninteractive {
        return true;
    }
    let mut output = std::io::stderr().lock();
    write!(output, "{msg} [y/N] ").unwrap();
    output.flush().unwrap();
    let mut user_input = String::new();
    std::io::stdin().read_line(&mut user_input).unwrap();
    user_input.trim().eq_ignore_ascii_case("y")
}

fn run_simulation(
    data_args: &OutputDataArgs,
    config: &Config,
    seed: Option<u64>,
    id: usize,
    tx: &mpsc::SyncSender<Msg>,
    exit_early: &atomic::AtomicBool,
) {
    // create lists to collect move/metric data into
    let mut moves = data_args.list_moves.then(Vec::new);
    let mut metrics = data_args.metrics.then(Vec::new);

    // initialize the cheese race simulation
    let seed = seed.unwrap_or_else(|| rand::thread_rng().next_u64());
    let mut sim = Simul::new(
        seed,
        SimulOptions {
            goal: config.game.goal,
            garbage: config.game.rules.min..=config.game.rules.max,
            previews: config.game.rules.previews,
        },
    );

    let time_start = Instant::now();

    'playing: while sim.lines_left() > 0 {
        if exit_early.load(atomic::Ordering::Relaxed) {
            break 'playing;
        }

        // initialize the bot for the current state
        let mut bot = Bot::new(
            &config.bot.weights,
            sim.matrix(),
            Queue::new(sim.queue().hold(), sim.queue().next()),
        );
        if data_args.metrics {
            bot.start_instrumenting();
        }

        // run the bot for a while; this could have been just `bot.think_for(iters)` but
        // we limit the gas argument to prevent the thread from hanging too long while
        // thinking.
        while bot.iterations() < config.bot.iters {
            let gas = std::cmp::min(50_000, config.bot.iters - bot.iterations());
            bot.think_for(gas);

            // send a heartbeat back to the main thread to test if its still alive
            if tx.send(Msg::Heartbeat(id)).is_err() {
                break 'playing;
            }
            if exit_early.load(atomic::Ordering::Relaxed) {
                break 'playing;
            }
        }

        let Some(placement) = bot.suggest() else {
            // this should never fail unless something went wrong e.g. not enough iters,
            // impossible board state, or buggy bot
            tracing::warn!("gave up :( pieces={}, id={id}", sim.pieces());
            break 'playing;
        };

        sim.play(placement);
        if let Some(moves) = moves.as_mut() {
            moves.push(Move { placement })
        }
        if let Some(metrics) = metrics.as_mut() {
            metrics.push(bot.metrics().unwrap());
        }

        let status = Status {
            lines_left: sim.lines_left(),
            pieces: sim.pieces(),
            time: Instant::now() - time_start,
        };
        if tx.send(Msg::Status(id, status.into())).is_err() {
            break 'playing;
        }
    }

    let pieces = sim.pieces();
    let cleared = config.game.goal - sim.lines_left();
    let output = Output {
        seed,
        config: config.game.clone(),
        cleared,
        pieces,
        moves,
        metrics,
    };
    if tx.send(Msg::Output(id, output.into())).is_err() {
        tracing::debug!("nobody received output id={id}");
    }
}

fn run_simulations(
    data_args: &OutputDataArgs,
    config: &Config,
    id: usize,
    tx: &mpsc::SyncSender<Msg>,
    tasks: &atomic::AtomicI64,
    exit_early: &atomic::AtomicBool,
) {
    while tasks.fetch_sub(1, atomic::Ordering::Relaxed) > 0 {
        if exit_early.load(atomic::Ordering::Relaxed) {
            break;
        }

        run_simulation(data_args, config, None, id, tx, exit_early);
    }

    if tx.send(Msg::Shutdown(id)).is_err() {
        tracing::debug!("nobody received shutdown id={id}");
    }
}
