use fish::bot::Metrics;
use mino::srs::Piece;
use serde::Serialize;

use crate::config::GameConfig;

#[derive(Serialize, Debug)]
pub struct Output {
    #[serde(serialize_with = "ser::seed")]
    pub seed: u64,
    pub config: GameConfig,
    pub pieces: usize,
    pub cleared: usize,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub moves: Option<Vec<Move>>,
    #[serde(
        serialize_with = "ser::metrics",
        skip_serializing_if = "Option::is_none"
    )]
    pub metrics: Option<Vec<Metrics>>,
}

impl Output {
    pub fn did_complete(&self) -> bool {
        self.cleared == self.config.goal
    }
}

#[derive(Serialize, Debug)]
pub struct Move {
    #[serde(flatten, serialize_with = "ser::placement")]
    pub placement: Piece,
    // TODO: spin?
}

#[derive(Copy, Clone, Debug)]
pub struct SummaryStats<T: Copy> {
    count: usize,
    min: Option<T>,
    max: Option<T>,
    sum: Option<T>,
}

impl<T: Copy> Default for SummaryStats<T> {
    fn default() -> Self {
        SummaryStats {
            count: 0,
            min: None,
            max: None,
            sum: None,
        }
    }
}

impl<T: Copy> SummaryStats<T> {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn count(&self) -> usize {
        self.count
    }

    pub fn min(&self) -> Option<T> {
        self.min
    }

    pub fn max(&self) -> Option<T> {
        self.max
    }
}

impl<T> SummaryStats<T>
where
    T: Ord + Copy + std::ops::Add<T, Output = T>,
{
    pub fn insert(&mut self, x: T) {
        self.count += 1;
        self.min = Some(self.min.map_or(x, |y| x.min(y)));
        self.max = Some(self.max.map_or(x, |y| x.max(y)));
        self.sum = Some(self.sum.map_or(x, |y| x + y));
    }
}

impl SummaryStats<f64> {
    pub fn insert_f64(&mut self, x: f64) {
        self.count += 1;
        self.min = Some(self.min.map_or(x, |y| x.min(y)));
        self.max = Some(self.max.map_or(x, |y| x.max(y)));
        self.sum = Some(self.sum.map_or(x, |y| x + y));
    }
}

impl<T> SummaryStats<T>
where
    T: Copy + std::ops::Div<usize, Output = T>,
{
    pub fn avg(&self) -> Option<T> {
        self.sum.map(|x| x / self.count)
    }
}

impl<T> SummaryStats<T>
where
    T: Copy,
    f64: From<T>,
{
    pub fn avg_f64(&self) -> Option<f64> {
        self.sum.map(|x| f64::from(x) / self.count as f64)
    }
}

impl<T> Serialize for SummaryStats<T>
where
    T: Serialize + Copy + std::ops::Div<usize, Output = T>,
{
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        ser::summary_stats(self, serializer)
    }
}

mod ser {
    use super::*;

    pub fn seed<S>(seed: &u64, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        // :')
        seed.to_string().serialize(serializer)
    }

    pub fn placement<S>(pc: &Piece, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        #[derive(Serialize)]
        struct Placement {
            #[serde(rename = "type")]
            ty: char,
            x: i16,
            y: i16,
            r: char,
        }

        Placement {
            ty: pc.ty.as_char(),
            x: pc.loc.x,
            y: pc.loc.y,
            r: match pc.loc.r {
                mino::Rot::N => 'N',
                mino::Rot::E => 'E',
                mino::Rot::S => 'S',
                mino::Rot::W => 'W',
            },
        }
        .serialize(serializer)
    }

    pub fn metrics<S>(metrics: &Option<Vec<Metrics>>, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        #[derive(Serialize)]
        struct MoveMetrics {
            start: NodeMetrics,
            end: NodeMetrics,
        }

        #[derive(Serialize)]
        struct NodeMetrics {
            features: Vec<i32>,
            heuristic: i32,
            #[serde(flatten)]
            rating: RatingVariant,
            #[serde(skip_serializing_if = "Option::is_none")]
            iteration: Option<u32>,
        }

        #[derive(Serialize)]
        #[serde(untagged)]
        enum RatingVariant {
            NotApplicable,
            Rating { rating: i32 },
            Solution { solution: i32 },
        }

        metrics
            .iter()
            .flat_map(|m| m.iter())
            .map(|m| MoveMetrics {
                start: NodeMetrics {
                    features: m.start_features.1.to_vec(),
                    heuristic: m.start_heuristic,
                    rating: RatingVariant::NotApplicable,
                    iteration: None,
                },
                end: NodeMetrics {
                    features: m.end_features.1.to_vec(),
                    heuristic: m.end_heuristic,
                    rating: match m.end_rating {
                        (false, v) => RatingVariant::Rating { rating: v },
                        (true, v) => RatingVariant::Solution { solution: -v },
                    },
                    iteration: Some(m.end_iteration),
                },
            })
            .collect::<Vec<_>>()
            .serialize(serializer)
    }

    pub fn summary_stats<S, T>(stats: &SummaryStats<T>, serializer: S) -> Result<S::Ok, S::Error>
    where
        T: Serialize + Copy + std::ops::Div<usize, Output = T>,
        S: serde::Serializer,
    {
        #[derive(Serialize)]
        struct Stats<T> {
            #[serde(skip_serializing_if = "Option::is_none")]
            min: Option<T>,
            #[serde(skip_serializing_if = "Option::is_none")]
            max: Option<T>,
            #[serde(skip_serializing_if = "Option::is_none")]
            avg: Option<T>,
        }

        Stats {
            min: stats.min,
            max: stats.max,
            avg: stats.avg(),
        }
        .serialize(serializer)
    }
}
